"""
This module contains the ResultList class that adds extra features to lists.

MIT License

Copyright (c) 2023 Joakim Wassberg

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation 
files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, 
copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom 
the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE 
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import functools
import random


class ResultList(list):
    """
    Extends the list class with methods to retrieve the first or last value,
    or None if the list is empty, and additional methods for filtering,
    mapping, reducing, sorting, grouping, and selecting a random element.
    This class is used as a return value for returned documents
    """

    def first_or_none(self):
        """
        Return the first value or None if list is empty
        :return: First list element or None
        """
        return self[0] if len(self) > 0 else None
    
    def first(self):
        """
        Return the first value or None if list is empty
        Synonym for first_or_none
        :return: First list element or None
        """
        return self.first_or_none()

    def last_or_none(self):
        """
       Return the last value or None if list is empty
       :return: Last list element or None
       """
        return self[-1] if len(self) > 0 else None

    def last(self):
        """
        Return the last value or None if list is empty
        Synonym for last_or_none
        :return: Last list element or None
        """
        return self.last_or_none()
    
    
    def filter(self, predicate):
        """
        Return a new ResultList containing only elements that match a given predicate function
        :param predicate: A function that takes an element and returns a boolean value
        :return: A new ResultList containing only matching elements
        """
        return ResultList(filter(predicate, self))

    def map(self, mapper):
        """
        Apply a given function to each element in the list and return a new ResultList containing the results
        :param mapper: A function that takes an element and returns a new value
        :return: A new ResultList containing the results of applying the mapper function to each element
        """
        return ResultList(map(mapper, self))

    def reduce(self, reducer, initial=None):
        """
        Reduce the list to a single value using a given reducer function
        :param reducer: A function that takes two elements and returns a single value
        :param initial: An optional initial value to start the reduction
        :return: The final reduced value
        """
        if initial is not None:
            return functools.reduce(reducer, self, initial)
        else:
            return functools.reduce(reducer, self)

    def sort(self, key=None, reverse=False):
        """
        Sort the list in place using a given sorting function
        :param key: A function that takes an element and returns a value to sort by
        :param reverse: A boolean indicating whether to sort in descending order (default is ascending)
        :return: None
        """
        super().sort(key=key, reverse=reverse)

    def group_by(self, keyfunc):
        """
        Group the elements in the list by a given key function and return a dictionary where the keys are the group keys
        and the values are lists of elements in that group.
        :param keyfunc: A function that takes an element and returns a key to group by
        :return: A dictionary of group keys and lists of elements in each group
        """
        groups = {}
        for elem in self:
            key = keyfunc(elem)
            if key in groups:
                groups[key].append(elem)
            else:
                groups[key] = [elem]
        return groups

    def random(self):
        """
        Return a random element from the list
        :return: A random element from the list
        """
        return None if len(self) < 1 else random.choice(self)



#!/usr/bin/env python

"""Tests for `mongeasy` package."""
# Do this before importing mongeasy as it will try to connect to the database
import os
os.environ['MONGOEASY_CONNECTION_STRING'] = 'mongodb://localhost:27017'
os.environ['MONGOEASY_DATABASE_NAME'] = 'mongeasy_pytest'
from pymongo import MongoClient
import pytest


from mongeasy import create_document_class


@pytest.fixture(scope='function')
def clean_mongo():
    # Connect to MongoDB
    client = MongoClient('mongodb://localhost:27017')
    db = client['mongeasy_pytest']

    # Clean all collections in the test database
    for collection_name in db.list_collection_names():
        db[collection_name].delete_many({})



def test_document_init_with_dict(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    stored_user = User.find({'first_name': 'John'}).first()
    assert user.first_name == stored_user.first_name
    assert user.last_name == stored_user.last_name
    assert user.age == stored_user.age


def test_document_init_with_kwargs(clean_mongo):
    User = create_document_class('User', 'users')
    user = User(first_name='Pete', last_name='Doe', age=30)
    user.save()
    stored_user = User.find({'first_name': 'Pete'}).first()
    assert user.first_name == stored_user.first_name
    assert user.last_name == stored_user.last_name
    assert user.age == stored_user.age
    assert True


# def test_document_init_with_positional_arguments(clean_mongo):
#     assert True


# def test_document_init_with_invalid_arguments(clean_mongo):
#     assert True


def test_document_has_changed_with_no_changes(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    assert not user.has_changed()


def test_document_has_changed_with_changes(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    user.age += 1
    assert user.has_changed()


def test_document_has_changed_with_new_document(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})

    assert user.has_changed()


def test_document_is_saved_with_no_changes(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    assert user.is_saved()

def test_document_is_saved_with_changes(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    user.age += 1
    assert not user.is_saved()


def test_document_save_new_document(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    stored_user = User.find({'first_name': 'John'}).first()
    assert user.first_name == stored_user.first_name
    assert user.last_name == stored_user.last_name
    assert user.age == stored_user.age


def test_document_save_existing_document_with_no_changes(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    user.save()
    stored_user = User.find({'first_name': 'John'}).first()
    assert user.first_name == stored_user.first_name
    assert user.last_name == stored_user.last_name
    assert user.age == stored_user.age


def test_document_save_existing_document_with_changes(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    user.age += 1
    user.save()
    stored_user = User.find({'first_name': 'John'}).first()
    assert user.first_name == stored_user.first_name
    assert user.last_name == stored_user.last_name
    assert user.age == stored_user.age


def test_document_reload_existing_document(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    user.age += 1
    user.reload()
    assert user.age == 30


def test_document_reload_unsaved_document(clean_mongo):
    from mongeasy.exceptions import MongEasyDBDocumentError
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    with pytest.raises(MongEasyDBDocumentError):
        user.reload()

def test_document_delete_field(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    user.delete_field('age')
    user.save()
    stored_user = User.find({'first_name': 'John'}).first()
    assert 'age' not in stored_user


def test_document_delete_existing_document(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    user.delete()
    stored_user = User.find({'first_name': 'John'}).first()
    assert stored_user is None


def test_document_delete_unsaved_document(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.delete()
    stored_user = User.find({'first_name': 'John'}).first()
    assert stored_user is None


def test_document_to_json(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    json_user = user.to_json()
    assert isinstance(json_user, str)
    assert 'John' in json_user
    assert 'Doe' in json_user
    assert '30' in json_user
    assert str(user._id) in json_user


def test_document_get_by_id_with_valid_id(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    stored_user = User.get_by_id(user._id)
    assert user.first_name == stored_user.first_name
    assert user.last_name == stored_user.last_name
    assert user.age == stored_user.age


def test_document_get_by_id_with_invalid_id(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    stored_user = User.get_by_id('123')
    assert stored_user is None


def test_document_insert_many(clean_mongo):
    users = [
    {'first_name': 'Alice', 'last_name': 'Smith', 'age': 25},
    {'first_name': 'Bob', 'last_name': 'Jones', 'age': 35},
    {'first_name': 'Charlie', 'last_name': 'Brown', 'age': 45},
    {'first_name': 'Dave', 'last_name': 'Smith', 'age': 55}
]
    User = create_document_class('User', 'users')
    User.insert_many(users)
    stored_users = User.all()
    assert len(stored_users) == 4

def test_document_find(clean_mongo):
    User = create_document_class('User', 'users')
    user = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user.save()
    stored_user = User.find({'first_name': 'John'}).first()
    assert user.first_name == stored_user.first_name
    assert user.last_name == stored_user.last_name
    assert user.age == stored_user.age


def test_document_find_in(clean_mongo):
    User = create_document_class('User', 'users')
    user1 = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user1.save()
    user2 = User({'first_name': 'Alice', 'last_name': 'Smith', 'age': 25})
    user2.save()
    user3 = User({'first_name': 'Bob', 'last_name': 'Jones', 'age': 35})
    user3.save()
    stored_users = User.find_in('first_name', ['John', 'Bob'])
    assert len(stored_users) == 2
    assert user1.first_name in [user.first_name for user in stored_users]
    assert user2.first_name not in [user.first_name for user in stored_users]
    assert user3.first_name in [user.first_name for user in stored_users]


def test_document_all(clean_mongo):
    User = create_document_class('User', 'users')
    user1 = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user1.save()
    user2 = User({'first_name': 'Alice', 'last_name': 'Smith', 'age': 25})
    user2.save()
    user3 = User({'first_name': 'Bob', 'last_name': 'Jones', 'age': 35})
    user3.save()
    stored_users = User.all()
    assert len(stored_users) == 3
    assert user1.first_name in [user.first_name for user in stored_users]
    assert user2.first_name in [user.first_name for user in stored_users]
    assert user3.first_name in [user.first_name for user in stored_users]


def test_document_delete(clean_mongo):
    User = create_document_class('User', 'users')
    user1 = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user1.save()
    user2 = User({'first_name': 'Alice', 'last_name': 'Smith', 'age': 25})
    user2.save()
    user3 = User({'first_name': 'Bob', 'last_name': 'Jones', 'age': 35})
    user3.save()
    User.delete({'first_name': 'Alice'})
    stored_users = User.all()
    assert len(stored_users) == 2
    assert user1.first_name in [user.first_name for user in stored_users]
    assert user2.first_name not in [user.first_name for user in stored_users]
    assert user3.first_name in [user.first_name for user in stored_users]


def test_document_document_count(clean_mongo):
    User = create_document_class('User', 'users')
    user1 = User({'first_name': 'John', 'last_name': 'Doe', 'age': 30})
    user1.save()
    user2 = User({'first_name': 'Alice', 'last_name': 'Smith', 'age': 25})
    user2.save()
    user3 = User({'first_name': 'Bob', 'last_name': 'Jones', 'age': 35})
    user3.save()
    count = User.document_count()
    assert count == 3




"""Top-level package for Mongeasy."""

__author__ = """Joakim Wassberg"""
__email__ = 'joakim.wassberg@arthead.se'
__version__ = '0.1.4'

import os
import configparser
import pymongo
from mongeasy.exceptions import MongEasyDBConnectionError
from mongeasy.dynamics import create_document_class

connection = None


def get_connection():
    return connection


def connect(connection_str: str=None, db_name: str=None):
    global connection
    if connection is not None:
        return connection

    try:
        # Try connecting using connect function
        if connection_str and db_name:
            connection = pymongo.MongoClient(connection_str)[db_name]
            return connection
    except Exception as e:
        print(e)
    
    connect_from_env()
    connect_from_config()

    if connection is None:
        raise MongEasyDBConnectionError("Failed to connect to database using environment variables or config file.")

    return connection

def connect_from_env():
    global connection
    try:
        # Try connecting using environment variables
        connection_string = os.environ.get('MONGOEASY_CONNECTION_STRING')
        database_name = os.environ.get('MONGOEASY_DATABASE_NAME')
        if connection_string and database_name:
            connection = pymongo.MongoClient(connection_string)[database_name]
    except Exception as e:
        print(e)

def connect_from_config():
    global connection
    try:
        # Try connecting using config file
        config = configparser.ConfigParser()
        config.read('mongeasy.conf')
        if 'mongoeasy' in config:
            section = config['mongoeasy']
            connection_string = section.get('connection_string')
            database_name = section.get('database_name')
            if connection_string and database_name:
                connection = pymongo.MongoClient(connection_string)[database_name]
    except Exception as e:
        print(e)
        
        
connect()